import fetch from 'node-fetch'
import crypto from 'crypto'
import { config } from './config.js'
import { getRedisKey, isJSFile } from './util.js'
import { getCSSDependenciesFor, getViteManifests } from './manifests.js'
import { client } from './redis.js'
import { logger } from './logger.js'
import { NotFoundError } from './errors.js'

const fileCache = {}

export async function fetchFileWithHeadersFromBaseUrl (path, baseUrl, version) {
  const [response, dependencies] = await Promise.all([
    fetch(new URL(path, baseUrl)),
    isJSFile(path) && getCSSDependenciesFor({ file: path.substr(1), version })
  ])

  if (!response.ok) throw new NotFoundError(`Error fetching file: ${path}`)

  const cssString = dependencies && dependencies.map(file => `"${file}"`).join(',')
  const appendix = cssString && `\n/*injected by ui-middleware*/document.dispatchEvent(new CustomEvent("load-css",{detail:{css:[${cssString}]}}))`
  const resBuffer = await response.arrayBuffer()
  const appendixLength = appendix?.length || 0
  const body = Buffer.alloc(resBuffer.byteLength + appendixLength)

  body.fill(Buffer.from(resBuffer), 0, resBuffer.byteLength)
  if (appendix) body.write(appendix, resBuffer.byteLength)

  const sha256Sum = crypto.createHash('sha256').update(body).digest('base64')

  return {
    body,
    sha256Sum,
    headers: {
      'content-type': response.headers.get('content-type'),
      dependencies
    }
  }
}

export async function fetchFileWithHeaders ({ path, version }) {
  if (config.urls.length === 0) await config.load()

  const viteManifests = await getViteManifests({ version })
  const module = viteManifests[path.substr(1)]
  if (module?.meta?.baseUrl) {
    try {
      return fetchFileWithHeadersFromBaseUrl(path, module.meta.baseUrl, version)
    } catch (err) {
      logger.debug(`File ${path} had a baseUrl but could not be found on that server: ${err}`)
    }
  }

  return Promise.any(config.urls.map(baseUrl => fetchFileWithHeadersFromBaseUrl(path, baseUrl, version)))
}

export async function saveToCache ({ version, path, body, headers, ...rest }) {
  if (typeof body !== 'string' && !(body instanceof Buffer)) body = JSON.stringify(body)
  fileCache[version] = fileCache[version] || {}
  fileCache[version][path] = {
    body,
    headers,
    ...rest
  }
  return Promise.all([
    client.set(getRedisKey({ version, name: `${path}:body` }), body),
    client.set(getRedisKey({ version, name: `${path}:meta` }), JSON.stringify({ headers, ...rest }))
  ])
}

export async function loadFromCache ({ version, path }) {
  if (!fileCache[version]?.[path]) {
    const [body, meta = '{}'] = await Promise.all([
      client.getBuffer(getRedisKey({ version, name: `${path}:body` })),
      client.get(getRedisKey({ version, name: `${path}:meta` }))
    ])
    if (!body) return
    fileCache[version] = fileCache[version] || {}
    fileCache[version][path] = { ...JSON.parse(meta), body }
  }
  return fileCache[version][path]
}
